
--Copyright (c) 1993 Ohio Board of Regents and the University of
--Cincinnati.  All Rights Reserved.

--author: sidhartha mohanty, philip a. wilsey.
--smohanty@thor.ece.uc.edu, phil.wilsey@uc.edu


library petri;
use petri.petri.all;
use std.textio.all;
entity place is

    generic (number_in_arcs  : integer := 1;
             number_out_arcs : integer := 1;
             initial_tokens  : integer := 0;
             pl_name         : string  := "";
             debug_control   : debug   := none);

    port (in_arc  : inout arc_array (1 to number_in_arcs) bus;
          out_arc : inout arc_array (1 to number_out_arcs) bus);
    
end;


architecture behavior of place is

--accumulation of tokens in place is held by three signals.  the
--tokens being sent in the current major cycle are stored in
--number_tokens_incoming.  the number of tokens in this place at the
--last major cycle boundary is stored in number_tokens_held.  finally,
--the number of tokens that have been taken from this place in the
--current major cycle are accumulated in number_tokens_outgoing.  when
--responding to transition requests for token information, the answer
--(tokens present/absent) is determined by subtracting the number of
--outgoing tokens from the number of held tokens.  the number of
--incoming tokens is not applied until the next major cycle boundary.

    signal number_tokens_incoming : integer := 0;
    signal number_tokens_held     : integer := initial_tokens;
    signal number_tokens_outgoing : integer := 0;


--unique identifier read from file produced by disambiguate is stored
--in identifer.

    signal identifier : integer := 1;
    disconnect all : arc after 0 ns;


--in_arc_list and out_arc_list store the identifier numbers from the
--places in, respectively, the input arcs and the output arcs.

    signal in_arc_list  : arc_list (1 to number_in_arcs);
    signal out_arc_list : arc_list (1 to number_out_arcs);

begin


--get disambiguating identifier from file produced by disambiguate.

initialize: process 

    variable integer_line : line;
    variable integer_id : integer := 1;

begin

    for i in 1 to number_in_arcs loop
	in_arc(i) <= null;
    end loop;

    for i in 1 to number_out_arcs loop
    	out_arc(i) <= null;
    end loop;

    wait for 0 fs;  -- let transitions get numbers first

    assert not std.textio.endfile(std.textio.input)
        report "(place.vhdl) Number of places and transitions exceed max_places_and_transitions.  Increase the value in package petri."
        severity failure;

    readline(std.textio.input, integer_line);

    assert integer_line'length >= 0 
        report "(place.vhdl) Missing integer in file used to disambiguate places and transitions.  Check to make certain that disambiguate was run properly."
        severity failure;

    read(integer_line, integer_id);
    identifier <= integer_id;
    
    if (debug_control = handshaking or
        debug_control = everything) then 
    	print_debug(id_a_place, debug_control, 
	            pl_name, integer_id,
                    "Identifying self to transitions.");
    end if;

    for i in 1 to number_in_arcs loop
    	in_arc(i) <= (identify_place, integer_id);
    end loop;

    for i in 1 to number_out_arcs loop
    	out_arc(i) <= (identify_place, integer_id);
    end loop;

    wait for 0 fs; 

    if (debug_control = handshaking or
        debug_control = everything) then 
    	print_debug(id_a_place, debug_control,
	            pl_name, identifier,
                    "Ending identification phase.");
    end if;

    for i in 1 to number_in_arcs loop
    	in_arc(i) <= null;
    end loop;

    for i in 1 to number_out_arcs loop
    	out_arc(i) <= null;
    end loop;

    wait for 0 fs;

    wait;

end process;


--locate transitions connected to my ports

find_transitions: process 

begin

    wait for 0 fs;

    while now < 1 ns loop

        wait on in_arc,out_arc;
	
    	for i in 1 to number_in_arcs loop

    	    if in_arc(i).handshake = identify_transition then

    	    	if debug_control = handshaking or
	       	   debug_control = everything then 
   	    	    print_debug(id_a_place, debug_control, 
	            	    	pl_name, identifier,
                    	    	"Received identify handshake from " &
			    	stringof(in_arc(i).number) & ".");
	    	end if;

	    	in_arc_list(i) <= in_arc(i).number;

    	    end if;

    	end loop;

    	for i in 1 to number_out_arcs loop

    	    if out_arc(i).handshake = identify_transition then

    	    	if debug_control = handshaking or
	       	   debug_control = everything then 
   	    	    print_debug(id_a_place, debug_control, 
	            	    	pl_name, identifier,
                    	    	"Received identify handshake from " &
			    	stringof(out_arc(i).number) & ".");
	    	end if;

	    	out_arc_list(i) <= out_arc(i).number;

    	    end if;

    	end loop;

    end loop;

   if debug_control = handshaking or
       debug_control = everything then 

    	for i in 1 to number_in_arcs loop
    	    if in_arc_list(i) = arc_list_element'left then
	    	print_debug(id_a_place, debug_control, 
	            	    pl_name, identifier,
                    	    "I identify input port " &
			    stringof(i) & 
			    " as being open.");
	    end if;
	end loop;

    	for i in 1 to number_out_arcs loop
    	    if out_arc_list(i) = arc_list_element'left then
	    	print_debug(id_a_place, debug_control, 
	            	    pl_name, identifier,
                   	    "I identify output port " &
			    stringof(i) & 
			    " as being open.");
	    end if;
	end loop;

    end if;

    for i in 1 to number_in_arcs loop
    	in_arc(i) <= null;
    end loop;

    for i in 1 to number_out_arcs loop
    	out_arc(i) <= null;
    end loop;

    wait for 0 fs;

    wait;  --end transition identification phase

end process;


--count input tokens sent from transitions during minor cycles
--(femtoseconds).  

count_incoming: process

    variable wk_time : time := 1 fs;
    variable i : integer := 1;
    variable temp : bool_array(1 to number_in_arcs);
begin

    wait for 1 ns;  --delay for initializing place and transition identifiers.

    loop

	 --compute  offset to next ns boundary and place in wk_time

    	i := (now / 1 ns);
    	wk_time := (1 ns * (i + 1)) - now;

	if (debug_control = everything) then
   	    print_debug(id_a_place, debug_control,
	                pl_name, identifier,
			"Distance to next ns is computed as " &
			stringof(wk_time) & "(incoming counts).");
    	end if;

        wait on in_arc for wk_time;


    	for i in 1 to number_in_arcs loop

	    if (in_arc(i).handshake = sending) then
	      temp(i) := true;
	    end if;
	end loop;

	for i in 1 to number_in_arcs loop
	  if temp(i) = true then
	    temp(i) := false;
	    number_tokens_incoming <= number_tokens_incoming + 1;
            wait for 0 fs;

	   if (debug_control = trace_tokens or
    	       debug_control = everything) then
    	       print_debug(id_a_place, debug_control,
	            	   pl_name, identifier,
			   "Receiving token from " &
			   stringof(in_arc(i).number) & ".");
	   end if;

	  end if ;

	end loop;

    	i := (now / 1 ns);
    	wk_time := 1 ns * i;

    	if (wk_time = now) then  --at a ns boundary?

	    number_tokens_incoming <= 0;  --no transitions will be
					  --active in these intervals
            wait for 0 fs;
	
	    if (debug_control = everything) then
    	    	print_debug(id_a_place, debug_control,
		            pl_name, identifier,
			    "Resetting incoming token counter.");
	    end if;

    	end if;

    end loop;

end process;


--update number of tokens held at major cycle boundary (nanoseconds).

update_token_count: process begin

    wait for 1 ns;  --delay for initializing place and transition identifiers.

    if (debug_control = trace_tokens or
        debug_control = everything) then
    	print_debug(id_a_place, debug_control,
	            pl_name, identifier,
		    "I am starting with an initial token marking of " &
		    stringof(number_tokens_held));
    end if;

    loop 

        wait for 1 ns;  --move to next major cycle boundary

    	if (debug_control = everything) then
    	    print_debug(id_a_place, debug_control,
	                pl_name, identifier,
			"Revising token counts.  " &
			"Held: " & stringof(number_tokens_held) &
			", incoming: " & stringof(number_tokens_incoming) &
			", outgoing: " & stringof(number_tokens_outgoing));

       end if;
	number_tokens_held <= number_tokens_held + number_tokens_incoming -
	                                           number_tokens_outgoing;
        wait for 0 fs;
    	if (debug_control = trace_tokens) then
    	    print_debug(id_a_place, debug_control,
	                pl_name, identifier,
			"I am now holding " &
			stringof(number_tokens_held) & " tokens.");
        end if;

    end loop;

end process;


--send token information to requesting transitions.

send_token_info: process 

    variable i : integer := 1;
    variable transition_number : integer := 1;
    variable number_tokens_requested : integer := 0;
    variable avail : bool_array(1 to number_out_arcs);
begin

    wait for 1 ns;  --delay for initializing place and transition identifiers.

    loop

	if (debug_control = handshaking or
	    debug_control = everything) then
    	    print_debug(id_a_place, debug_control,
	                pl_name, identifier,
			"Waiting for a transition to request token information.");
        end if;

    	wait on out_arc;

	i := 1;

	while i <= number_out_arcs and 
	    out_arc(i).handshake /= send loop
	    i := i + 1;
	end loop;
	
	if i <= number_out_arcs then

	    transition_number := out_arc(i).number;
	    number_tokens_requested := 1;
            i := i + 1; -- ??????

	    while i <= number_out_arcs loop

	        if out_arc(i).handshake = send then

		    if out_arc(i).number = transition_number then

		    	number_tokens_requested := number_tokens_requested + 1;

		    else

        	    	print_debug(id_a_place, debug_control,
    	                            pl_name, identifier,
    			            "Illegal request for token information " &
                                    "from multiple transitions (" &
				    stringof(transition_number) & "," &
				    stringof(out_arc(i).number) & ").");
        	    	assert false severity failure;

		    end if;

		end if;

		i := i + 1;

	    end loop;

	    if (debug_control = handshaking or
    	        debug_control = everything) then
        	print_debug(id_a_place, debug_control,
    	                    pl_name, identifier,
    			    "Received request for token information from: " &
   			    stringof(transition_number));
            end if;
    
--replace next if statement with the following code if you want to
--update place information at every minor cycle boundary.

--    	  if (number_tokens_held +
--    	        number_tokens_incoming -
--    	        number_tokens_outgoing -
--		number_tokens_requested >= 0) then

    	    if (number_tokens_held -
    	        number_tokens_outgoing -
		number_tokens_requested >= 0) then
    
    	    	    if (debug_control = handshaking or
   	            	debug_control = everything) then
        	    	print_debug(id_a_place, debug_control,
    	                            pl_name, identifier,
                                    "Telling Transition that I have the " &
				    stringof(number_tokens_requested) &
				    " token(s) requested.");
                    end if;
    
    	    	for i in 1 to number_out_arcs loop
		    if out_arc(i).handshake = send then
		       avail(i) := TRUE;
		    end if;
		end loop;
	
		for i in 1 to number_out_arcs loop
		       out_arc(i) <= null;
		end loop;

		wait for 0 fs;

		for i in 1 to number_out_arcs loop
		    if avail(i) = TRUE then
		       avail(i) := FALSE;
    	    	       out_arc(i) <= (present, identifier);
		    end if;
		end loop;

	        wait for 0 fs;

	        for i in 1 to number_out_arcs loop
	            if out_arc(i).handshake = present then
			out_arc(i) <= null;
		    end if; 
		end loop;

		wait for 0 fs;

	    else
    	    	if (debug_control = handshaking or
    	            debug_control = everything) then
        	    print_debug(id_a_place, debug_control,
    		                pl_name, identifier,
                                "Telling Transition that I do not have a token.");
                end if;
    
	    end if;
    
          end if;

    end loop;

end process;


--updating outgoing token count.

count_outgoing: process

    variable wk_time : time;
    variable i : integer;
    variable transition_firing : integer;
    variable number_firing : integer;

begin

    wait for 1 ns;  --delay for initializing place and transition identifiers.

    loop

	 --compute  offset to next ns boundary and place in wk_time

    	i := (now / 1 ns);
    	wk_time := (1 ns * (i + 1)) - now;

	if (debug_control = everything) then
    	    print_debug(id_a_place, debug_control,
	                pl_name, identifier,
			"Distance to next ns is computed as " &
			stringof(wk_time) & "(outgoing counts).");
    	end if;

	wait on out_arc for wk_time;

	transition_firing := 0;
    	number_firing := 0;
    	
    	for i in 1 to number_out_arcs loop
	    
	    if (out_arc(i).handshake = firing) then

		out_arc(i) <= null;
	        number_firing := number_firing + 1;

		if transition_firing = 0 then
		    transition_firing := out_arc(i).number;
		elsif transition_firing /= out_arc(i).number then
        	    print_debug(id_a_place, debug_control,
    	                        pl_name, identifier,
    			        "Illegal firing of multiple transitions (" &
				stringof(transition_firing) & "," &
				stringof(out_arc(i).number) & ").");
        	    assert false severity failure;
		end if;
	      
	    end if;

	end loop;

	wait for 0 fs;
	
	if number_firing > 0 then

	    number_tokens_outgoing <= number_tokens_outgoing + number_firing;
	    wait for 0 fs;
         
	    if (debug_control = trace_tokens or
   	        debug_control = everything) then
    	    	print_debug(id_a_place, debug_control,
		            pl_name, identifier,
			    stringof(number_firing) &
			    " token were sent to " &
                           stringof(transition_firing) & ".");
	    end if;

	end if;

    	i := (now / 1 ns);
    	wk_time := 1 ns * i;

    	if (wk_time = now) then  --at a ns boundary?

	    number_tokens_outgoing <= 0;  --no transitions will be
					  --active in these intervals
	    wait for 0 fs;
	    if (debug_control = everything) then
    	    	print_debug(id_a_place, debug_control,
		            pl_name, identifier,
			    "Resetting outgoing token counter.");
	    end if;

    	end if;

    end loop;

end process;

end behavior;
